﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.IO;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using System.Text.RegularExpressions;

namespace ScanSort
{
    public partial class Form1 : Form
    {
        public Form1()
        {
            InitializeComponent();
        }

        // ========================================================================================

        private void Form1_Load(object sender, EventArgs e)
        {
            LoadScanFolder();
        }

        // ========================================================================================

        private void SelectScanFolderToolStripMenuItem_Click(object sender, EventArgs e)
        {
            vistaFolderBrowserDialog1.SelectedPath = GetPref("activeFolder");
            if (vistaFolderBrowserDialog1.ShowDialog() == DialogResult.OK)
            {
                SetPref("activeFolder", vistaFolderBrowserDialog1.SelectedPath);
                LoadScanFolder();
            }
        }

        private void RefreshToolStripMenuItem_Click(object sender, EventArgs e)
        {
            LoadScanFolder();
        }

        private void AboutToolStripMenuItem_Click(object sender, EventArgs e) => new Form2().ShowDialog();

        private void ExitToolStripMenuItem_Click(object sender, EventArgs e) => this.Close();

        // ========================================================================================

        public string GetPref(string key)
        {
            return Properties.Settings.Default[key].ToString();
        }

        public void SetPref(string key, string value)
        {
            Properties.Settings.Default[key] = value;
            Properties.Settings.Default.Save();
        }

        public string[] GetDir(string path, bool getFiles)
        {
            string[] result = { };
            try
            {
                result = getFiles ? Directory.GetFiles(path) : Directory.GetDirectories(path);
            }
            catch (Exception e)
            {
                MessageBox.Show(e.ToString(), "Could not access folder", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
            return result;
        }

        // ========================================================================================

        public void HideImage()
        {
            button1.Enabled = false;
            button2.Enabled = false;
            listBox1.Enabled = false;
            pictureBox1.Image = null;
        }

        private void LoadImage(string fileName)
        {
            pictureBox1.Image = Image.FromStream(new MemoryStream(File.ReadAllBytes(fileName)));
            button1.Enabled = true;
            button2.Enabled = true;
            listBox1.Enabled = true;
        }

        private void RotateImage(RotateFlipType rotateFlipType)
        {
            Cursor.Current = Cursors.WaitCursor;
            pictureBox1.Image.RotateFlip(rotateFlipType);
            pictureBox1.Image.Save((string)treeView1.SelectedNode.Tag, System.Drawing.Imaging.ImageFormat.Jpeg);
            pictureBox1.Refresh();
            Cursor.Current = Cursors.Default;
        }

        private void Button1_Click(object sender, EventArgs e)
        {
            RotateImage(RotateFlipType.Rotate270FlipNone);
            treeView1.Focus();
        }

        private void Button2_Click(object sender, EventArgs e)
        {
            RotateImage(RotateFlipType.Rotate90FlipNone);
            treeView1.Focus();
        }

        // ========================================================================================

        public void LoadScanFolder()
        {
            if (GetPref("activeFolder").Length == 0) SetPref("activeFolder", Environment.GetFolderPath(Environment.SpecialFolder.Desktop));

            toolStripStatusLabel1.Text = GetPref("activeFolder");

            treeView1.BeginUpdate();
            treeView1.Nodes.Clear();

            listBox1.BeginUpdate();
            listBox1.Items.Clear();
            listBox1.Items.Add("Uncategorized");


            // Load uncategorized files
            TreeNode uncategorizedNode = treeView1.Nodes.Add("Uncategorized");
            uncategorizedNode.Tag = "Uncategorized";
            uncategorizedNode.ImageIndex = 0;
            uncategorizedNode.SelectedImageIndex = 0;
            string[] uncategorizedFiles = GetDir(GetPref("activeFolder"), true);
            int uncategorizedFileCounter = 0;
            foreach (string uncategorizedFile in uncategorizedFiles)
            {
                string extension = Path.GetExtension(uncategorizedFile).ToLower();
                if (extension == ".jpg" || extension == ".jpeg")
                {
                    TreeNode newNode = uncategorizedNode.Nodes.Add(Path.GetFileName(uncategorizedFile));
                    newNode.Tag = uncategorizedFile;
                    newNode.ImageIndex = 1;
                    newNode.SelectedImageIndex = 1;
                    uncategorizedFileCounter++;
                }
            }
            uncategorizedNode.Text += " (" + uncategorizedFileCounter + ")";


            // Load first-level subfolders as categories
            foreach (string categoryFolder in GetDir(GetPref("activeFolder"), false))
            {
                string categoryName = Path.GetFileName(categoryFolder);
                listBox1.Items.Add(categoryName);

                // Load categorized files
                TreeNode categoryNode = treeView1.Nodes.Add(categoryName);
                categoryNode.Tag = categoryName;
                categoryNode.ImageIndex = 0;
                categoryNode.SelectedImageIndex = 0;
                int categoryFileCounter = 0;
                foreach (string categoryFile in GetDir(categoryFolder, true))
                {
                    string extension = Path.GetExtension(categoryFile).ToLower();
                    if (extension == ".jpg" || extension == ".jpeg")
                    {
                        TreeNode imageNode = categoryNode.Nodes.Add(Path.GetFileName(categoryFile));
                        imageNode.Tag = categoryFile;
                        imageNode.ImageIndex = 1;
                        imageNode.SelectedImageIndex = 1;
                        categoryFileCounter++;
                    }
                }
                categoryNode.Text += " (" + categoryFileCounter + ")";
            }

            listBox1.EndUpdate();
            treeView1.EndUpdate();

            SelectActiveNode();
        }

        // ========================================================================================

        public void SelectActiveNode()
        {
            TreeNode[] matchingNodes = treeView1.FlattenTree().Where(n => (string)n.Tag == GetPref("activeNode")).ToArray();
            if (matchingNodes.Length > 0)
            {
                treeView1.SelectedNode = matchingNodes[0];
            }
            else
            {
                SetPref("activeNode", "Uncategorized");
                treeView1.SelectedNode = treeView1.Nodes[0];
            }
            treeView1.SelectedNode.EnsureVisible();
            treeView1.Focus();
        }

        public void SelectPreviousNode()
        {
            if (GetPref("previousNode").Length == 0)
            {
                SetPref("previousNode", GetPref("activeNode"));
                treeView1.Focus();
            }
            else
            {
                string previousNode = GetPref("activeNode");
                SetPref("activeNode", GetPref("previousNode"));
                SelectActiveNode();
                SetPref("previousNode", previousNode);
            }
        }

        private void Button6_Click(object sender, EventArgs e)
        {
            SelectPreviousNode();
        }

        public void CollapseNodes()
        {
            string previousNode = GetPref("activeNode");
            if (treeView1.SelectedNode.Level > 0)
            {
                SetPref("activeNode", (string)treeView1.SelectedNode.Parent.Tag);
            }
            treeView1.CollapseAll();
            SelectActiveNode();
            SetPref("previousNode", previousNode);
        }

        public void ExpandNodes()
        {
            treeView1.ExpandAll();
            SelectActiveNode();
        }

        private void Button8_Click(object sender, EventArgs e)
        {
            CollapseNodes();
        }

        private void Button7_Click(object sender, EventArgs e)
        {
            ExpandNodes();
        }

        // ========================================================================================

        private void TreeView1_KeyDown(object sender, KeyEventArgs e)
        {
            if (e.Alt || e.Control) return;
            switch (e.KeyCode)
            {
                case Keys.Space: case Keys.Enter: if (treeView1.SelectedNode.IsExpanded) treeView1.SelectedNode.Collapse(); else treeView1.SelectedNode.Expand(); break;
                case Keys.Escape: SelectPreviousNode(); break;
                case Keys.F1: CollapseNodes(); break;
                case Keys.F2: ExpandNodes(); break;
                case Keys.F3: if (listBox1.Enabled && listBox1.SelectedIndex > 0) listBox1.SelectedIndex--; break;
                case Keys.F4: if (listBox1.Enabled && listBox1.SelectedIndex + 1 < listBox1.Items.Count) listBox1.SelectedIndex++; break;
                case Keys.F5: if (MakeCategoryPath(maskedTextBox1.Text.Trim()) == "") maskedTextBox1.Focus(); else NewCategory(maskedTextBox1.Text.Trim()); break;
                case Keys.F6: RotateImage(RotateFlipType.Rotate270FlipNone); break;
                case Keys.F7: RotateImage(RotateFlipType.Rotate90FlipNone); break;
                default: return;
            }
            e.Handled = true;
        }

        private void TreeView1_KeyPress(object sender, KeyPressEventArgs e)
        {
            e.Handled = true;
        }

        private void TreeView1_AfterSelect(object sender, TreeViewEventArgs e)
        {
            SetPref("previousNode", GetPref("activeNode"));
            SetPref("activeNode", (string)treeView1.SelectedNode.Tag);

            string fileName = (string)treeView1.SelectedNode.Tag;
            string categoryName;
            if (treeView1.SelectedNode.Level == 0)
            {
                // Category node is selected
                categoryName = GetPref("activeNode");
                button3.Text = "<F5>: &New";
                HideImage();
            }
            else
            {
                // Image node is selected
                categoryName = Path.GetFileName(Path.GetDirectoryName(fileName));
                button3.Text = "<F5>: Move to &New";
                LoadImage(fileName);
            }

            // Synchronize category to selected node
            int matchingIndex = listBox1.FindStringExact(categoryName);
            listBox1.SelectedIndex = matchingIndex > -1 ? matchingIndex : 0;

            button4.Enabled = listBox1.Enabled && listBox1.SelectedIndex > 0;
            button5.Enabled = listBox1.Enabled && listBox1.SelectedIndex < listBox1.Items.Count - 1;
        }

        // ========================================================================================

        private void MoveFile(string sourceFileName, string destCategory)
        {
            string newFileName = GetPref("activeFolder") + "\\" + (destCategory == "Uncategorized" ? "" : destCategory + "\\") + Path.GetFileName(sourceFileName);
            if (newFileName != sourceFileName)
            {
                string previousNode = GetPref("previousNode");
                File.Move(sourceFileName, newFileName);
                SetPref("activeNode", newFileName);
                LoadScanFolder();
                SetPref("previousNode", previousNode);
            }
        }

        private void ListBox1_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (treeView1.SelectedNode.Level > 0)
            {
                MoveFile((string)treeView1.SelectedNode.Tag, listBox1.SelectedItem.ToString());
            }

            button4.Enabled = listBox1.Enabled && listBox1.SelectedIndex > 0;
            button5.Enabled = listBox1.Enabled && listBox1.SelectedIndex < listBox1.Items.Count - 1;
        }

        private void Button4_Click(object sender, EventArgs e)
        {
            if (listBox1.Enabled && listBox1.SelectedIndex > 0) listBox1.SelectedIndex--;
        }

        private void Button5_Click(object sender, EventArgs e)
        {
            if (listBox1.Enabled && listBox1.SelectedIndex + 1 < listBox1.Items.Count) listBox1.SelectedIndex++;
        }

        // ========================================================================================

        private void MaskedTextBox1_KeyPress(object sender, KeyPressEventArgs e)
        {
            if (e.KeyChar == (char) Keys.Enter)
            {
                string categoryName = maskedTextBox1.Text.Trim();
                if (categoryName.Length > 0)
                {
                    NewCategory(categoryName);
                    e.Handled = true;
                }
            }
        }

        private void Button3_Click(object sender, EventArgs e)
        {
            string categoryName = maskedTextBox1.Text.Trim();
            if (categoryName.Length > 0) NewCategory(categoryName);
        }

        public string MakeCategoryPath(string categoryName)
        {
            if (categoryName.Length == 0 || Regex.Matches(categoryName, @"[/\\]").Count > 0) return "";
            try
            {
                return Path.GetFullPath(GetPref("activeFolder") + "\\" + categoryName);
            }
            catch (Exception)
            {
                return "";
            }
        }

        public void NewCategory(string categoryName)
        {
            string newPath = MakeCategoryPath(categoryName);
            if (newPath == "")
            {
                MessageBox.Show("Please avoid special characters", "Could not create Category Subfolder", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
                return;
            }

            maskedTextBox1.Clear();

            TreeNode selectedNode = treeView1.SelectedNode;

            Directory.CreateDirectory(newPath);

            if (selectedNode.Level == 0)
            {
                SetPref("activeNode", categoryName);
                LoadScanFolder();
            }
            else
            {
                MoveFile((string)treeView1.SelectedNode.Tag, categoryName);
            }
        }

        private void MaskedTextBox1_TextChanged(object sender, EventArgs e)
        {
            button3.Enabled = MakeCategoryPath(maskedTextBox1.Text.Trim()) != "";
        }

        // ========================================================================================
    }

    public static class SOExtension
    {
        public static IEnumerable<TreeNode> FlattenTree(this TreeView tv) => FlattenTree(tv.Nodes);

        public static IEnumerable<TreeNode> FlattenTree(this TreeNodeCollection coll) => coll.Cast<TreeNode>().Concat(coll.Cast<TreeNode>().SelectMany(x => FlattenTree(x.Nodes)));
    }

}
